//
// SOME CONFIGURATION VARIABLES
//
// High and very query times will be emphasized in the UI.
var high_query_time_ms = 200;
var very_high_query_time_ms = 1000;
// Caching files is faster, but makes debugging harder.
var cache_files = false;
// Threshold beyond which a position in the suggestion list counts as infinite.
var infinite_position = 1000000;
// Page size for computing MRR (the parameter k from the paper).
var mrr_page_size = 7;
var dir_regex = "";
var result_regex = "";

// THIS CODE WORKS AS FOLLOWS:
//
// 0. Assume that python3 -m http.server <port> was started in <directory>
// 1. It looks for all folders output... in <directory>
// 2. For a selected output folder, it looks for all <name>.results.tsv
// 3. For a selected result file, it displays it as a nice table and also
//    reads the correspondins <name>.completion-queries.yml
// 4. A click on a row in the table shows the corresponding SPARQL query
//    and the execution tree, which is drawn using the TREANT library
//
// TODO: The YAML could be improved in various ways: (1) write the operations in
// UPPER CASE; (2) shorten the variable names, write them in lowercase and give
// them more sensible names, like ?subject or ?object or ?name. Some of these
// improvements are currently done in this script, see the function
// runtimeInfoForTreant
//

// Global variables.
var details = {}
var target_queries = {}
// Remember the last_scroll_position in the result table TAB.
var result_table_element_last_clicked_on = undefined;
var result_table_last_scroll_position = 0;

$(document).ready(function() {

  //
  // INITIALIZATION
  //
  $.ajaxSetup({ cache: cache_files, ifModified:true });
  $("#results").tabs({
    activate: function(event, ui) {
      var tab_index = ui.newTab.index();
      window.location.hash = "tab-" + tab_index;
      if (tab_index == 0 && result_table_element_last_clicked_on != undefined) {
        // $(result_table_element_last_clicked_on).closest("tr").css("background-color", "#DDDDAA");
        $(result_table_element_last_clicked_on).scrollTop(result_table_last_scroll_position);
        // console.log($(result_table_element_last_clicked_on).closest("tr").css("background-color"));
      }
    }
  });

  // For switching between tabs using the browswer back and forward buttons.
  $.address.change(function(event){
    // console.log("hash: " + window.location.hash);
    $("#results").tabs("option", "active",  window.location.hash.replace("#tab-", ""));
  })
  // when the tab is selected update the url with the hash

  $("#select-output-dir").selectmenu();
  $("#select-result-file").selectmenu();
  var base_path = window.location.pathname.replace(/www.*/, "results/");
  console.log("Base Path: " + base_path);
 
  // Parse query parameters and set the corresponding variables.
  if (window.location.search.startsWith("?")) {
    params = {}
    window.location.search.substring(1).split("&").forEach(function(pair_string) {
      pair = pair_string.split("=");
      if (pair.length == 2) { params[pair[0]] = pair[1]; }
    })
    if ("mrr_page_size" in params) { mrr_page_size = parseInt(params["mrr_page_size"]); }
    if ("dir" in params) { dir_regex = params["dir"]; }
    if ("res" in params) { result_regex = params["res"]; }
  }

  //
  // GET AVAILABLE OUTPUT DIRECTORIES
  //
  // This works by parsing the directory listing provided by Python's
  // http.server
  //
  $.get(base_path, function(data) {
    var output_directories = Array()
    $(data).find("a").each(function(){
      var name = $(this).attr("href");
      if (name.match(dir_regex)) {
        output_directories.push(name);
        $("#output-options").append(new Option(name, name));
      }
    });
    output_directories.sort();
    output_directories.reverse();
    $("#output-options").val(output_directories[0]);
    $("#output-options").trigger("change");
  });

  //
  // GET AVAILABLE RESULT FILES.
  //
  // When an output directory has been selected, parse the directory in a
  // similar fashion as described above.
  //
  $("#output-options").change(function(){
    output_directory = $("#output-options option:selected").val();
    console.log("output_directory: ");
    console.log(output_directory);
    $("#result-options option").remove();
    $.get(base_path + output_directory, function(data) {
      var result_files = Array()
      $(data).find("a").each(function(){
        var name = $(this).attr("href");
        if (name.match(".results.tsv$") && name.match(result_regex)) {
          result_files.push(name);
          $("#result-options").append(new Option(name, name));
        }
      });
      $("#result-options").val(result_files[0]);
      $("#result-options").trigger("change");
    })
  })



  //
  // READ TSV AND YAML FILE AND SHOW TABLE.
  //
  // The table is constructed manually (it's not that hard). It is made sortable using
  // the Tablesorter plugin for jQuery. Rows with large query times are
  // highlighted (class .high for >= high_query_time_ms, class .veryhigh for >=
  // very_high_query_time_ms, see the CSS file).
  //
  // Also, basic statistics of the query times are shown (average, media,
  // 90%-ile and maximum).
  //
  // The YAML file is parsed using JS-YAML. The resulting JSON is rearranged
  // such that we can access the SPARQL queries and Runtime Info via the
  // completion query key (e.g. "alan-turing-fields-of-work-completion-1")
  //
  $("#result-options").change(function(){
    
    // Get paths to the various files.
    var output_directory = $("#output-options option:selected").val();
    var result_file = $("#result-options option:selected").val();
    var result_path = output_directory + result_file;
    var queries_path = result_path.replace(".completion.results.tsv", ".input-queries.yml");
    var details_path = result_path.replace(".results.tsv", ".executed.yml");

    // Signal in the statistics line that we are computing ...
    $("#result-statistics-1").html("Computing ...");
    $("#result-statistics-1").css("color", "darkred");
    $("#result-statistics-1").css("font-weight", "bold");
    $("#result-statistics-2").html("");
    $("#result-statistics-3").html("");
    $("#result-table-div").hide();

    // Read and process the corresponding .input-queries.yml file.
    target_queries = {}
    console.log("Reading target queries from \"" + queries_path + "\"");
    var target_query_promise = new Promise(function(resolve, reject) {
      $.get(base_path + queries_path, function(data) {
        data = jsyaml.safeLoadAll(data);
        if (data.length == 1) { data = data[0]; }
        var queries_as_array = data["queries"];

        queries_as_array.forEach(function(item) {
          target_queries[item["query"]] = item["sparql"];
        })
        console.log("Target queries read, #queries = ", Object.keys(target_queries).length);
        resolve();
      })
    });

    // Read and process the corresponding .result.tsv file. Make sure that the
    // target queries have already been computed (used for the tooltips).
    target_query_promise.then(function() {
      console.log("Reading results file: ", result_path);
      $.get(base_path + result_path, function(data) {
        // Get lines of the result (omit empty part after last \n).
        var lines = data.split("\n").slice(0, -1);
        console.log("Results file read, #lines = ", lines.length);
        // Create table.
        create_detailed_results_table(lines);
        $("#result-table-div").show();
      })
    });
 
    // Read the corresponding .completion-queries.yml file.
    console.log("Reading runtime info from \"" +  details_path + "\"");
    details = {}
    $.get(base_path + details_path, function(data) {
      // For some reason, safeLoadAll sometimes returns an array with a single
      // element, which in turn is the array of all documents, hence the [0].
      data = jsyaml.safeLoadAll(data);
      if (data.length == 1) { data = data[0]; }
      var details_as_array = data;

      details_as_array.forEach(function(item) {
        details[item["query"]] = {
          sparql : item["sparql"],
          runtime_info: item["runtime_info"]
        };
      });
      console.log("Runtime info read, #records = ", Object.keys(details).length);
    });
  });
});


// Create table with detailed results from given lines from results.tsv file.
//
// TODO: this is a monster function doing too many things at the same time.
// Using the classes from statistics.js will alread reduce it somewhat.
function create_detailed_results_table(lines) {

  // Compute the following statistics
  //
  // TODO: use the classes in statistics.js for this.
  var query_times = [];
  var query_times_per_token_warm = [];
  var query_times_per_token_cold = [];
  var query_times_per_token_context = [];
  var query_times_per_token_no_context = [];
  var num_failed_queries_warm = 0;
  var num_failed_queries_cold = 0;
  var num_failed_queries_context = 0;
  var num_failed_queries_no_context = 0;
  var num_failed_tokens_warm = 0;
  var num_failed_tokens_cold = 0;
  var num_failed_tokens_context = 0;
  var num_failed_tokens_no_context = 0;

  var table = "<table class=\"tablesorter\">\n";
  // Which columns to show and consider in the following.
  var query_name_column = 1;
  var query_time_column = 7;
  var partial_triple_column = 3;
  var alias_column = 4;
  var triple_num_colum = 5;
  var position_column = 9;
  var keystrokes_column = 10;
  var client_time_column = 11;
  var has_context_column = 6;
  var is_warmup_query_column = 12;
  var is_new_token_column = 13;
  var was_fallback_query_executed_column = 14;
  var is_dummy_query_column = 15;
  var has_failed_column = 16;
  var target_cert_failed_column = 18;
  var page_cert_failed_column = 19;
  var context_warmup_intersection_column= 334;
  var context_warmup_intersection= 0;
  // var is_new_token_computed_column = -1;

  // Show the following columns in the table.
  var columns =
    [query_name_column, partial_triple_column, alias_column,
     query_time_column, /* client_time_column, */ position_column,
     keystrokes_column, has_context_column, /*is_warmup_query_column,*/
     is_new_token_column, has_failed_column, /*was_fallback_query_executed_column*/
     target_cert_failed_column, page_cert_failed_column];
     // /* , is_dummy_query_column, context_warmup_intersection_column */];
  var max_colum_index = columns.reduce((x, y) => Math.max(x, y));

  // Keep track of some things.
  var stats_per_query = {};
  var queries_with_large_ks = {};
  var mrr_histogram = {};   // All tokens
  var mrrw_histogram = {};  // All tokens warmup
  var mrrc_histogram = {};  // All tokens with context
  var mrrnw_histogram = {}; // All tokens without warmup
  var mrrnc_histogram = {}; // All tokens without context
  var mrrw_sum = 0;
  var mrrc_sum = 0;
  var mrrnw_sum = 0;
  var mrrnc_sum = 0;
  var num_tokens_warm = 0;
  var num_tokens_cold = 0;
  var num_tokens_context = 0;
  var num_tokens_no_context = 0;
  var has_current_token_failed = 0;
  var ks_histogram = {};
  var previous_partial_triple = "";
  var previous_prefix_length = 0;

  console.log("Iterating over results file ...");

  // Iterate over all lines from the results file.
  //
  // NOTE: This part of the code is not the problem performance-wise, it is
  // reasonably fast.
  $.each(lines, function(i, value) {
    var line = lines[i].split("\t")

    // Parse values
    var target_query_name;
    var completion_query_name;
    var partial_triple;
    var triple_num;
    var query_time;
    var client_time;
    var position;
    var ks;
    var has_context;
    var is_warmup_query;
    var is_new_token;
    var was_fallback_query_executed;
    var is_dummy_query;
    var has_failed;
    var target_cert_failed;
    var page_cert_failed;
    // var is_new_token_computed;

    // Update statistics.
    if (i > 0) {

      is_new_token = parseInt(line[is_new_token_column]);
      was_fallback_query_executed = was_fallback_query_executed_column < line.length
                                      ? parseInt(line[was_fallback_query_executed_column]) : 0;
      is_dummy_query = is_dummy_query_column < line.length
          ? parseInt(line[is_dummy_query_column]) : 0;


      // HACK 16.09.2020: If was_fallback_query_executed, the query time
      // should be at least the timeout (currently 2 seconds).
      //
      // TODO: this should be fixed in the eval.py script
      /*
      if (was_fallback_query_executed && is_new_token && parseInt(line[query_time_column]) < 2000)
        line[query_time_column] = "2000.00";
        */

      // Parse values.
      //
      // TODO: in the evaluation script, remove trailing whitespace. It is
      // currently removed here explicitly, but that is bound to lead to
      // bugs.
      target_query_name = line[0];
      completion_query_name = line[query_name_column];
      partial_triple = line[partial_triple_column].replace(/\s*$/, ""); 
      triple_num = line[triple_num_colum];
      query_time = line[query_time_column] != "" ? parseInt(line[query_time_column]) : Number.MAX_VALUE;
      client_time = line[client_time_column] != "" ? parseInt(line[client_time_column]) : "None";
      position = parseInt(line[position_column]); if (position >= infinite_position) position = Number.MAX_VALUE;
      ks = parseInt(line[keystrokes_column]);
      has_context = parseInt(line[has_context_column]);
      is_warmup_query = parseInt(line[is_warmup_query_column]);
      context_warmup_intersection = has_context==is_warmup_query;

      has_failed = has_failed_column < line.length ? parseInt(line[has_failed_column]) : query_time == Number.MAX_VALUE;
      target_cert_failed = target_cert_failed_column < line.length ? parseInt(line[target_cert_failed_column]) : " ";
      page_cert_failed = page_cert_failed_column < line.length ? parseInt(line[page_cert_failed_column]) : " ";

      // Check if we are the beginning of a new target query.
      if (!(target_query_name in stats_per_query)) {
        previous_partial_triple = "";
        stats_per_query[target_query_name] = {
          "num_completion_queries": 0,
          "num_tokens" : 0,
          "mrr_per_token": [],
          "ks_per_token": []
        };
        // console.log(stats_per_query[query_name]);
      }

      // Check if this is the beginning of a new token (by checking if
      // deletion of the last character before the cursor is equal to the
      // previous partial triple).
      // is_new_token_computed = partial_triple.replace(/.(\W)$/, "$1") != previous_partial_triple ? 1 : 0;
      // console.log("CHECK (" + i + "): \"" + partial_triple.replace(/.(\W)$/, "$1") + "\" != \"" + previous_partial_triple + "\" ?  -->  " + is_new_token);
      // line.push(is_new_token_computed.toString());

      // Some bookkeeping.
      previous_partial_triple = partial_triple;
      if (is_new_token) { has_current_token_failed = 0; }
      if (query_time != Number.MAX_VALUE) {
        query_times.push(query_time);
        // For the per-token query times (warm and cold), maintain the maximum
        var query_times_per_token = is_warmup_query ? query_times_per_token_warm : query_times_per_token_cold;
        var query_times_per_token_c = has_context ? query_times_per_token_context : query_times_per_token_no_context;
        if (is_new_token) {
          query_times_per_token.push(0);
          query_times_per_token_c.push(0);
        }
        if (query_time > query_times_per_token[query_times_per_token.length - 1]) {
          query_times_per_token[query_times_per_token.length - 1] = query_time; }
        if (query_time > query_times_per_token_c[query_times_per_token_c.length - 1]) {
          query_times_per_token_c[query_times_per_token_c.length - 1] = query_time; }
      } else {
        if (is_warmup_query) num_failed_queries_warm += 1;
        else num_failed_queries_cold += 1;
        if (has_context) num_failed_queries_context += 1;
        else num_failed_queries_no_context += 1;
        if (has_current_token_failed == 0) {
          has_current_token_failed = 1;
          if (is_warmup_query) num_failed_tokens_warm += 1;
          else num_failed_tokens_cold += 1;
          if (has_context) num_failed_tokens_context += 1;
          else num_failed_tokens_no_context += 1;
        }
      }

      // Keep track of the various statistics (only when the query did not
      // fail).
      if (query_time != -1) {

        // Update counts.
        stat = stats_per_query[target_query_name];
        stat["num_completion_queries"] += 1;
        if (is_new_token) stat["num_tokens"] += 1;
        // Update MRR array.
        if (is_new_token && !is_dummy_query) {
          // The position if infinite if the completion query has failed or if
          // the desired token is not in the top-k results.
          //
          // Also count a token as not suggested if it could not be certified in
          // agnostic mode (in that case, the eval script sets the query time to
          // infinity and hence has_current_token_failed will be true).
          var is_token_suggested = position < Number.MAX_VALUE && !has_current_token_failed;
          var rank = is_token_suggested ? 1 + Math.floor((position - 1) / mrr_page_size) : "\u221E";
          var reciprocal_rank = is_token_suggested ? 1 / rank : 0;
          stat["mrr_per_token"].push(reciprocal_rank);
          mrr_histogram[rank] = (mrr_histogram[rank] || 0) + 1;
          if (is_warmup_query == 1) {
            mrrw_histogram[rank] = (mrrw_histogram[rank] || 0) + 1;
            mrrw_sum += reciprocal_rank;
            num_tokens_warm += 1
          } else {
            mrrnw_histogram[rank] = (mrrnw_histogram[rank] || 0) + 1;
            mrrnw_sum += reciprocal_rank;
            num_tokens_cold += 1
          }
          if (has_context == 1) {
            mrrc_histogram[rank] = (mrrc_histogram[rank] || 0) + 1;
            mrrc_sum += reciprocal_rank;
            num_tokens_context += 1
          } else {
            mrrnc_histogram[rank] = (mrrnc_histogram[rank] || 0) + 1;
            mrrnc_sum += reciprocal_rank;
            num_tokens_no_context += 1
          }
        }
        // Update KS array (note that array assignment is by reference).
        var ks_array = stat["ks_per_token"];
        if (is_new_token) { ks_array.push(0); }
        ks_array[ks_array.length - 1] = ks;
        // Warning if large KS value
        if (ks > 10) { queries_with_large_ks[target_query_name] = ks; }
      }
    }

    // Create the line in the HTML table (except for the last dummy line).
    if (line[0] != "") {
      if (i == 0) { table += "<thead>\n"; }
      table += is_new_token ? "<tr class=\"new_token\">" : "<tr>";
      var tag = i == 0 ? "th" : "td";
      var cell_contents = i == 0 ? "#" : i.toString();
      table += "<" + tag + ">" + cell_contents +  "</" + tag + ">";
         
      // Highlight entries in certain cells: large query time, rank ∞ with
      // context at beginning of token
      var query_time_class = "";
      var position_class = "";
      var query_id_class = "";
      if (query_time >= very_high_query_time_ms || has_failed) query_time_class = " class=veryhigh";
      // if (query_time >= very_high_query_time_ms) query_time_class = " class=veryhigh";
      else if (query_time >= high_query_time_ms) query_time_class = " class=high";
      if (has_failed && is_new_token && has_context) position_class = " class=veryhigh";
      // if (position == Number.MAX_VALUE && is_new_token && has_context) position_class = " class=veryhigh";
      if (query_time_class.match(/=high/) || position_class.match(/=high/)) query_id_class = " class=high";
      if (query_time_class.match(/=very/) || position_class.match(/=very/)) query_id_class = " class=veryhigh";

      $.each(columns, function(j, value) {
        var col = columns[j];
        // Skip columns not available in earlier evaluations. Note that
        // "return" in a $.each loop is the equivalent of "continue" in a
        // normal for loop.

        var cell_contents = 0;

        if (col == context_warmup_intersection_column) {
          cell_contents = context_warmup_intersection.toString();
        } else {
          if (col >= line.length) return;
          var cell_contents = line[col];
        }
        var table_cell_attributes = "";

        if (i > 0) {

          // Some reformatting so that things look prettiert and correct.
          cell_contents = cell_contents.replace("-completion-", " ");
          cell_contents = cell_contents.replace(/</g, "&lt;");
          cell_contents = cell_contents.replace(/>/g, "&gt;");
          if (col == position_column && position == Number.MAX_VALUE) cell_contents = "\u221E";
          if (col == query_time_column && cell_contents == "") cell_contents = "failed";
          // if (col == query_time_column && query_time == Number.MAX_VALUE) cell_contents = "failed";
          // if (col == partial_triple_column) cell_contents = "[" + triple_num + "] " + cell_contents;
          if (col == alias_column && cell_contents.length > 15) {
            cell_contents = cell_contents.substr(0, 15) + "..."; }
          if (col == partial_triple_column && partial_triple.match(/\S{30}/)) {
            table_cell_attributes = " style=\"word-break: break-all\""; }

          // Add highlight info.
          var class_attribute = "";
          if (col == query_name_column) class_attribute = query_id_class;
          if (col == query_time_column) class_attribute = query_time_class;
          if (col == position_column) class_attribute = position_class;
          cell_contents = "<span " + class_attribute + ">" + cell_contents + "</span>";

          // Add tooltip for first column.
          var tooltip = "";
          if (col == query_name_column) tooltip = " title=\"" +
            target_queries[target_query_name].replace(/^[\s\S]*?SELECT/m, "SELECT").replace(/"/g, "'") + "\"";

          // Add link to other tab
          var tab_index = -1;
          if (col == query_name_column) tab_index = 1;
          if (col == partial_triple_column) tab_index = 2;
          if (col == query_time_column) tab_index = 3;
          if (tab_index != -1) {
            var href = "javascript:showDetailsForQuery(\"" +
              completion_query_name.replace(/\"/g, "\\\"").replace(/'/g, "&#39;") + "\", %J%, this, %I%)";
            cell_contents = "<a href='"
              + href.replace("%J%", tab_index)
                    .replace("%I%", was_fallback_query_executed)
              + "'" + tooltip + ">" + cell_contents + "</a>";
          }
        }
        table += "<" + tag + table_cell_attributes + ">" + cell_contents +  "</" + tag + ">";
      })
      table += "</tr>\n";
      if (i == 0) { table += "</thead>\n<tbody>\n"; }
    }
  })
  table += "</tbody></table>\n";
  $("#result-table table").html(table);
  $("#result-table table").tablesorter({sortStable : true, sortInitialOrder: "desc"});
  // Without the updateAll, sorting will not work anymore once the table is
  // replaced by a new one because tablesorter caches the contents.
  $("#result-table table").trigger("updateAll");
  $("#results").tabs("option", "active", "0");
  $("#result-table-div p span:eq(0)").hide();
  $("#result-table-div p span:eq(1)").show();

  // Identify rows with large and very large query times.
  // $("#result-table table a").
  //   filter(function(){return $(this).html() >= high_query_time_ms}).
  //   parent().parent().children().eq(1).addClass("high");
  // $("#result-table table a").
  //   filter(function(){return $(this).html() >= very_high_query_time_ms}).
  //   parent().parent().children().eq(1).addClass("veryhigh");

  console.log("Computing statistics ...");

  // Compute statistics for query times (mean, median, max, ...)
  //
  query_times = query_times.sort((x, y) => x - y);
  var n = query_times.length;
  var query_times_mean = query_times.reduce((prefix_sum, x) => prefix_sum + x) / n;
  var query_times_median = query_times[Math.round(0.50 * n)];
  var query_times_80perc = query_times[Math.round(0.80 * n)];
  var query_times_90perc = query_times[Math.round(0.90 * n)];
  var query_times_95perc = query_times[Math.round(0.95 * n)];
  var query_times_99perc = query_times[Math.round(0.99 * n)];
  var query_times_max = query_times[n - 1];

  var query_times_warm = query_times_per_token_warm.sort((x, y) => x - y);
  var nw = query_times_warm.length;
  var query_times_warm_mean = query_times_warm.reduce((prefix_sum, x) => prefix_sum + x, 0) / nw;
  var median_index_warm = Math.round(0.5 * (nw + num_failed_queries_warm));
  var query_times_warm_median = median_index_warm < query_times_warm.length
                                 ? Math.round(query_times_warm[median_index_warm]) : "\u221E";
  var query_times_warm_max = query_times_warm[nw - 1];
  var perc_query_time_warm_200ms = query_times_warm.filter(x => x <= high_query_time_ms).length / (nw + num_failed_queries_warm);
  var perc_query_time_warm_1000ms = query_times_warm.filter(x => x <= very_high_query_time_ms).length / (nw + num_failed_queries_warm);

  var query_times_cold = query_times_per_token_cold.sort((x, y) => x - y);
  var nc = query_times_cold.length;
  var query_times_cold_mean = query_times_cold.reduce((prefix_sum, x) => prefix_sum + x) / nc;
  var median_index_cold = Math.round(0.5 * (nc + num_failed_queries_cold));
  var query_times_cold_median = median_index_cold < query_times_cold.length
                                  ? Math.round(query_times_cold[median_index_cold]) : "\u221E";
  var query_times_cold_max = query_times_cold[nc - 1];
  var perc_query_time_cold_200ms = query_times_cold.filter(x => x <= high_query_time_ms).length / (nc + num_failed_queries_cold);
  var perc_query_time_cold_1000ms = query_times_cold.filter(x => x <= very_high_query_time_ms).length / (nc + num_failed_queries_cold);

  var query_times_context = query_times_per_token_context.sort((x, y) => x - y);
  var ncontext = query_times_context.length;
  var query_times_context_mean = query_times_context.reduce((prefix_sum, x) => prefix_sum + x) / ncontext;
  var median_index_context = Math.round(0.5 * (ncontext + num_failed_queries_context));
  var query_times_context_median = median_index_context < query_times_context.length
      ? Math.round(query_times_context[median_index_context]) : "\u221E";
  var query_times_context_max = query_times_context[ncontext - 1];
  var perc_query_time_context_200ms = query_times_context.filter(x => x <= high_query_time_ms).length / (ncontext + num_failed_queries_context);
  var perc_query_time_context_1000ms = query_times_context.filter(x => x <= very_high_query_time_ms).length / (ncontext + num_failed_queries_context);

  var query_times_no_context = query_times_per_token_no_context.sort((x, y) => x - y);
  var nno_context = query_times_no_context.length;
  var query_times_no_context_mean = query_times_no_context.reduce((prefix_sum, x) => prefix_sum + x, 0) / nno_context;
  var median_index_no_context = Math.round(0.5 * (nno_context + num_failed_queries_no_context));
  var query_times_no_context_median = median_index_no_context < query_times_no_context.length
      ? Math.round(query_times_no_context[median_index_no_context]) : "\u221E";
  var query_times_no_context_max = query_times_no_context[nno_context - 1];
  var perc_query_time_no_context_200ms = query_times_no_context.filter(x => x <= high_query_time_ms).length / (nno_context + num_failed_queries_no_context);
  var perc_query_time_no_context_1000ms = query_times_no_context.filter(x => x <= very_high_query_time_ms).length / (nno_context + num_failed_queries_no_context);
  // console.log(query_times_warm);
  // console.log(query_times_cold);

  // Compute statistics for MRR and KS
  // 
  var num_target_queries = 0;
  var num_tokens = 0;
  var mrr_sum = 0;
  var mrrp_sum = 0;
  var ks_sum = 0;
  var ks_max = 0;
  var mrrp_histogram = {};
  var ks_histogram = {};
  // console.log("Cats, #qrs : " + stats_per_query["Cats"]["num_completion_queries"]);
  // console.log("Cats, #tok : " + stats_per_query["Cats"]["num_tokens"]);
  // console.log("Cats, mrr  : " + stats_per_query["Cats"]["mrr_per_token"]);
  // console.log("Cats, ks   : " + stats_per_query["Cats"]["ks_per_token"]);
  // console.log("Goats, #qrs : " + stats_per_query["Goats"]["num_completion_queries"]);
  // console.log("Goats, #tok : " + stats_per_query["Goats"]["num_tokens"]);
  // console.log("Goats, mrr  : " + stats_per_query["Goats"]["mrr_per_token"]);
  // console.log("Goats, ks   : " + stats_per_query["Goats"]["ks_per_token"]);
  for (var target_query_name in stats_per_query) {
    // console.log(query_name);
    var stat = stats_per_query[target_query_name];
    var num_tokens_target_query = stat["num_tokens"];
    var mrr_sum_target_query = stat["mrr_per_token"].reduce((x, y) => x + y, 0);
    var mrr_avg_target_query = mrr_sum_target_query / num_tokens_target_query;
    var mrrp_avg_target_query =
      (mrr_sum_target_query - stat["mrr_per_token"][0]) / (num_tokens_target_query - 1);
    var ks_avg_target_query =
      stat["ks_per_token"].reduce((x, y) => x + y, 0) / num_tokens_target_query;
    mrr_sum += mrr_avg_target_query;
    mrrp_sum += mrrp_avg_target_query;
    ks_sum += ks_avg_target_query;
    num_target_queries += 1;
    num_tokens += num_tokens_target_query;
    // Update KS histogram (note that this cannot be computed while
    // building the table like the MRR histogram because for KS we only
    // now the final value for a token *after* we have seen the last
    // completion query for that token; the code above handles this
    // gracefully by simply overwriting the value until a new token is
    // started).
    $.each(stat["ks_per_token"], function(i, ks) {
      ks_histogram[ks] = (ks_histogram[ks] || 0) + 1;
      if (ks > ks_max) ks_max = ks;
    });
  }

  // Show some additional info on the console.
  // console.log("Queries with large KS: ");
  // $.each(queries_with_large_ks, function(target_query_name, ks) {
  //   console.log(target_query_name, ": KS = ", ks);
  // })
  // console.log("MRR Histogram: ", mrr_histogram)
  // console.log("MRRw Histogram: ", mrrw_histogram)
  // console.log("MRRnw Histogram: ", mrrnw_histogram)
  // console.log("MRRc Histogram: ", mrrc_histogram)
  // console.log("MRRnc Histogram: ", mrrnc_histogram)
  // console.log("KS Histogram: ", ks_histogram)
 
  // STATISTICS 1: number of queries of the various kinds.
  var statistics_1 =
    "# target queries = " + num_target_queries + ", " +
    "# tokens = " + num_tokens + ", " +
    "# completion queries = " + (n + num_failed_queries_warm + num_failed_queries_cold) + ", " +
    "# failed = " + (num_failed_queries_warm + num_failed_queries_cold) + " " +
    "... \u2300 KS<sub>" + mrr_page_size + "</sub>: " +
        (ks_sum / num_target_queries).toFixed(2) +
    " (max KS = " + ks_max + ")";
    // "Histogram KS: " + JSON.stringify(ks_histogram);
  $("#result-statistics-1").html(statistics_1);
  $("#result-statistics-1").css("color", "black");
  $("#result-statistics-1").css("font-weight", "normal");

  // Statistics warmup queries.
  // var statistics_2 =
  //   "# warm tokens (failed) = " + num_tokens_warm + " (" + num_failed_tokens_warm + "), " +
  //   "Med: " + query_times_warm_median + "ms, " + 
  //   "Avg" + (num_failed_queries_warm > 0 ? "': " : ": ") + Math.round(query_times_warm_mean) + "ms, " +
  //   "Max" + (num_failed_queries_warm > 0 ? "': " : ": ") + Math.round(query_times_warm_max) + "ms, " + 
  //   Math.round(100 * perc_query_time_warm_200ms) + "% < 200ms, " + 
  //   Math.round(100 * perc_query_time_warm_1000ms) + "% < 1s " +
  //   "... \u2300 MRRw<sub>" + mrr_page_size + "</sub>: " +
  //       Math.round(100 * mrrw_sum / num_tokens_warm) + "% " +
  //   "<small>[1: " + Math.round(100 * (mrrw_histogram[1] || 0) / num_tokens_warm) + "%, " +
  //   "2: " + Math.round(100 * (mrrw_histogram[2] || 0) / num_tokens_warm) + "%, ..., " +
  //   "\u221E: " + Math.round(100 * (mrrw_histogram["\u221E"] || 0) / num_tokens_warm) + "%]</small>";
  // $("#result-statistics-2").html(statistics_2);

  // Statistics per token without warmup
  // var statistics_3 =
  //   "# cold tokens (failed) = " + num_tokens_cold + " (" + num_failed_tokens_cold + "), " +
  //   "Med: " + query_times_cold_median + "ms, " + 
  //   "Avg" + (num_failed_queries_cold > 0 ? "': " : ": ") + Math.round(query_times_cold_mean) + "ms, " +
  //   "Max" + (num_failed_queries_cold > 0 ? "': " : ": ") + Math.round(query_times_cold_max) + "ms, " + 
  //   Math.round(100 * perc_query_time_cold_200ms) + "% < 200ms, " + 
  //   Math.round(100 * perc_query_time_cold_1000ms) + "% < 1s " +
  //   "... \u2300 MRRc<sub>" + mrr_page_size + "</sub>: " +
  //       Math.round(100 * mrrnw_sum / num_tokens_cold) + "% " +
  //   "<small>[1: " + Math.round(100 * (mrrnw_histogram[1] || 0) / num_tokens_cold) + "%, " +
  //   "2: " + Math.round(100 * (mrrnw_histogram[2] || 0) / num_tokens_cold) + "%, ..., " +
  //   "\u221E: " + Math.round(100 * (mrrnw_histogram["\u221E"] || 0) / num_tokens_cold) + "%]</small>";
  // $("#result-statistics-3").html(statistics_3);

  // Statistics per token with context
  var statistics_2 =
      "# context tokens (# failed) = " + num_tokens_context + " (" + num_failed_tokens_context + "), " +
      "Med: " + query_times_context_median + "ms, " +
      "Avg" + (num_failed_queries_context > 0 ? "': " : ": ") + Math.round(query_times_context_mean) + "ms, " +
      "Max" + (num_failed_queries_context > 0 ? "': " : ": ") + Math.round(query_times_context_max) + "ms, " +
      Math.round(100 * perc_query_time_context_200ms) + "% < 200ms, " +
      Math.round(100 * perc_query_time_context_1000ms) + "% < 1s " +
      "... \u2300 MRRc<sub>" + mrr_page_size + "</sub>: " +
      Math.round(100 * mrrc_sum / num_tokens_context) + "% " +
      "<small>[1: " + Math.round(100 * (mrrc_histogram[1] || 0) / num_tokens_context) + "%, " +
      "2: " + Math.round(100 * (mrrc_histogram[2] || 0) / num_tokens_context) + "%, ..., " +
      "\u221E: " + Math.round(100 * (mrrc_histogram["\u221E"] || 0) / num_tokens_context) + "%]</small>";
  $("#result-statistics-2").html(statistics_2);

  // Statistics per token without context
  var statistics_3 =
      "# context-free tokens (# failed) = " + num_tokens_no_context + " (" + num_failed_tokens_no_context + "), " +
      "Med: " + query_times_no_context_median + "ms, " +
      "Avg" + (num_failed_queries_no_context > 0 ? "': " : ": ") + Math.round(query_times_no_context_mean) + "ms, " +
      "Max" + (num_failed_queries_no_context > 0 ? "': " : ": ") + Math.round(query_times_no_context_max) + "ms, " +
      Math.round(100 * perc_query_time_no_context_200ms) + "% < 200ms, " +
      Math.round(100 * perc_query_time_no_context_1000ms) + "% < 1s " +
      "... \u2300 MRRc<sub>" + mrr_page_size + "</sub>: " +
      Math.round(100 * mrrnc_sum / num_tokens_no_context) + "% " +
      "<small>[1: " + Math.round(100 * (mrrnc_histogram[1] || 0) / num_tokens_no_context) + "%, " +
      "2: " + Math.round(100 * (mrrnc_histogram[2] || 0) / num_tokens_no_context) + "%, ..., " +
      "\u221E: " + Math.round(100 * (mrrnc_histogram["\u221E"] || 0) / num_tokens_no_context) + "%]</small>";
  $("#result-statistics-3").html(statistics_3);
}


//
// SHOW DETAILS OF PARTICULAR COMPLETION QUERY.
//
// Shows the SPARQL query and the Execution Tree (using Treant). In the
// execution tree, nodes with slow query times are highlighted according to thte
// same rules as in the table (see above).
//
// NOTE: It is important that the TAB is openened *before* the call to Treant,
// otherwise Treant assumes a too small div size and the whole drawing gets
// collapsed in an unreadable manner. This took me over an hour to find out.
//
// The element_from_result_table is the DOM element from the result table that
// was clicked on to trigger this function call. We use it to remember the
// scrolling position in the result table.
//
// NEW 16.09.2020: For mixed mode, now two SPARQL queries are executed. The last
// argument "query_index" says which of the two we should take (used as array
// index, that is, 0 or 1).
//
function showDetailsForQuery(query, tab_index, element_from_result_table, query_index) {
  // console.log("Details for: " + query);
  // console.log(details[query]);
 
  // Remember last scroll position and unhighlight previously highlighted row.
  // if (result_table_element_last_clicked_on != undefined) {
  //   $(result_table_element_last_clicked_on).closest("tr").css("background-color", "#DDDDDD"); }
  result_table_element_last_clicked_on = element_from_result_table;
  result_table_last_scroll_position = $(element_from_result_table).scrollTop();

  // Show SPARQL target query.
  var target_query = query.replace(/-completion-\d+.*/, "");
  $("#result-sparql-1-div p span:eq(0)").hide();
  $("#result-sparql-1-div p span:eq(1)").html("Target query for <b>" + target_query + "</b>");
  $("#result-sparql-1-div p span:eq(1)").show();
  $("#sparql-area-1")
    .html(document.createTextNode(target_queries[target_query]))
    .attr('rows', target_queries[target_query].split('\n').length);

  // Show SPARQL completion query.
  var query_time = details[query].runtime_info["total_time"];
  var query_html = "<b>" + query.replace(/-completion-(\d+)$/, " $1") +
                   "</b> (query time: " + query_time + "ms)";
  $("#result-sparql-2-div p span:eq(0)").hide();
  $("#result-sparql-2-div p span:eq(1)").html("Completion query for " + query_html);
  $("#result-sparql-2-div p span:eq(1)").show();
  // NEW 16.09.2020: For mixed mode, there are now two SPARQL queries (first
  // query = sensitive, second query = agnostic).
  // console.log("! Query index: " + query_index);
  if (Array.isArray(details[query].sparql)) {
    details[query].sparql = details[query].sparql[query_index];
  }
  // console.log("! SPARQL Query: " + details[query].sparql);
  $("#sparql-area-2")
    .html(document.createTextNode(details[query].sparql))
    .attr('rows', details[query].sparql.split('\n').length);

  // Show execution tree using TREANT.
  $("#result-tree-div p span:eq(0)").hide();
  $("#result-tree-div p span:eq(1)").html("Execution tree for " + query_html);
  $("#result-tree-div p span:eq(1)").show();
  $("#result-tree-div p span:eq(2)").show();
  var runtime_info = details[query].runtime_info;
  var have_runtime_info = Object.keys(runtime_info).length > 0;


  // Switch to tab specified by second argument.
  $("#results").tabs("option", "active", tab_index);
  // $("#results").tabs("option", "active", have_runtime_info ? "3" : "1");
  if (have_runtime_info) {
    runtimeInfoForTreant(runtime_info);
    var treant_tree = {
      chart: {
        container: "#result-tree",
        rootOrientation: "NORTH",
        connectors: { type: "step" },
      },
      nodeStructure: runtime_info
    }
    var treant_chart = new Treant(treant_tree);

    // Highlight node with high query times: cached -> yellow or light yellow, not
    // cached -> red or light red. Also grey out cached nodes.
    $("p.node-time").
      filter(function(){ return $(this).html() >= high_query_time_ms }).
      parent().addClass("high");
    $("p.node-time").
      filter(function(){ return $(this).html() >= very_high_query_time_ms }).
      parent().addClass("veryhigh");
    $("p.node-cached").
      filter(function(){ return $(this).html() == "true" }).
      parent().addClass("cached");
  }
}

//
// CONVERT RUNTIME INFO INTO FORMAT SUITABLE FOR TREANT.
//
// The runtime_info is a dictionary representing the execution tree. For drawing
// with Treant, this dictionary needs the form { children: [...], text: {...} },
// where [...] is an array of dictionaries of the same form (empty for the
// leaves) and {...} is the information of a node.
//
// The second argument is used to propagate the "cached" flag to all nodes of a
// subtree (in the drawing, it makes more sense to show all nodes in a subtree
// as cached, not just the root of the subtree).
//
// Returns sum of times from the children. This is needed so that we can compute
// the original operation time for cached nodes (original_(operation_)time
// stores the total time for the subtree).
//
function runtimeInfoForTreant(runtime_info, parent_cached = false) {

  // Create text child with the information we want to see in the tree.
  if (runtime_info["text"] == undefined) {
    var text = {};
    if (runtime_info["column_names"] == undefined) { runtime_info["column_names"] = ["not yet available"]; }
    text["name"] = runtime_info["description"]
                    .replace(/<.*[#\/\.](.*)>/, "<$1>")
                    .replace(/qlc_/g, "")
                    .replace(/\?[A-Z_]*/g, function(match) { return match.toLowerCase(); })
                    .replace(/([a-z])([A-Z])/g, "$1-$2")
                    .replace(/^([a-zA-Z-])*/, function(match) { return match.toUpperCase(); })
                    .replace(/([A-Z])-([A-Z])/g, "$1 $2")
                    .replace(/AVAILABLE /, "").replace(/a all/, "all");
    text["size"] = format(runtime_info["result_rows"]) + " x " + format(runtime_info["result_cols"]);
    text["cols"] = runtime_info["column_names"].join(", ")
                    .replace(/qlc_/g, "")
                    .replace(/\?[A-Z_]*/g, function(match) { return match.toLowerCase(); });
    text["time"] = runtime_info["was_cached"]
                    ? runtime_info["details"]["original_operation_time"]
                    : runtime_info["operation_time"];
    text["total"] = text["time"];
    text["cached"] = parent_cached == true ? true : runtime_info["was_cached"];
    // Save the original was_cached flag, before it's deleted, for use below.
    for (var key in runtime_info) { if (key != "children") { delete runtime_info[key]; } }
    runtime_info["text"] = text;
    runtime_info["stackChildren"] = true;

    // Recurse over all children, propagating the was_cached flag from the
    // original runtime_info to all nodes in the subtree.
    runtime_info["children"].map(child => runtimeInfoForTreant(child, text["cached"]));
    // If result is cached, subtract time from children, to get the original
    // operation time (instead of the original time for the whole subtree).
    if (text["cached"]) {
      runtime_info["children"].forEach(function(child) {
        // text["time"] -= child["text"]["total"];
      })
    }
  }
}

//
// FORMAT NUMBER USING THOUSAND SEPARATORS
//
// E.g., write 26362998243 as 26,362,998,243. Do it with a simple regex.
//
function format(number) {
  return number.toString().replace(/(\d)(?=(\d{3})+(?!\d))/g, "$1,");
}
